import re
import unittest
from time import sleep
from unittest import mock

from jacktrade import CodeTimer
from jacktrade.benchmark import *


class CodeTimerTest(unittest.TestCase):
    """
    Tests the CodeTimer class.
    """

    @staticmethod
    def flaky(test_method):
        """
        Marks the test as flaky, and suppresses AssertionError
        when running Python <= 3.10 on Windows.

        Prior to Python 3.11, the accuracy of sleep() on Windows was
        much lower, and millisecond accuracy is not reliably attainable.
        https://docs.python.org/3.11/library/time.html#time.sleep
        """

        def wrapper(self: unittest.TestCase, *args, **kwargs):
            try:
                test_method(self, *args, **kwargs)
            except AssertionError as ex:
                import platform
                import sys

                # Exemption for Python <= 3.10 running on Windows
                if sys.version_info[:2] <= (3, 10) and platform.system() == "Windows":
                    self.skipTest(
                        "sleep() function is insufficiently accurate on this platform"
                    )
                else:
                    raise ex  # Fail in all other cases

        return wrapper

    @flaky
    def test_core_functionality(self):
        """Tests the core functionality."""
        task_time_s = 0.100
        with CodeTimer() as ct:
            sleep(task_time_s)
        self.assertAlmostEqual(ct.s, task_time_s, 2)
        self.assertAlmostEqual(ct.ms / 1e3, task_time_s, 2)
        self.assertAlmostEqual(ct.us / 1e6, task_time_s, 2)
        self.assertAlmostEqual(ct.ns / 1e9, task_time_s, 2)

    @mock.patch("time.perf_counter_ns")
    def test_resolution(self, mock_perf_counter_ns):
        """Tests that calls are made to the high-resolution nanosecond performance counter."""
        mock_perf_counter_ns.side_effect = [0, 9]
        with CodeTimer() as ct:
            pass
        mock_perf_counter_ns.assert_called()

    @mock.patch("time.perf_counter_ns")
    def test_resolution(self, mock_perf_counter_ns):
        """No error must be raised if time delta is zero."""
        mock_perf_counter_ns.side_effect = [0, 0]
        with CodeTimer() as ct:
            pass
        # Test passes if no exceptions are raised

    @mock.patch("builtins.print")
    def test_print_options(self, mock_print):
        """Tests the no_print option."""
        with CodeTimer(no_print=True) as ct:
            pass
        mock_print.assert_not_called()
        with CodeTimer(no_print=False) as ct:
            pass
        mock_print.assert_called()

    def test_min_digits(self):
        """Tests that the min_digits parameter works as expected."""
        for time_ns in [
            int("".join(str(x % 10) for x in range(mx))) for mx in range(2, 22)
        ]:
            # [1, 12, 123, ... 12345678901234567890]
            for min_digits in range(1, 25):
                with self.subTest(time_ns=time_ns, min_digits=min_digits):
                    digit_count = sum(
                        c.isdigit() for c in CodeTimer._format_time(time_ns, min_digits)
                    )
                    self.assertGreaterEqual(digit_count, min_digits)

    def test__format_time(self):
        """Tests that time formatting works as expected."""
        self.assertEqual(
            CodeTimer._format_time(1, 3),
            "1.00 ns",
        )
        self.assertEqual(
            CodeTimer._format_time(NS_PER_MICROSECOND, 3),
            "1.00 us",
        )
        self.assertEqual(
            CodeTimer._format_time(NS_PER_MILLISECOND, 3),
            "1.00 ms",
        )
        self.assertEqual(
            CodeTimer._format_time(NS_PER_SECOND, 3),
            "1.00 s",
        )
        self.assertEqual(
            CodeTimer._format_time(NS_PER_MINUTE - NS_PER_MILLISECOND, 7),
            "59.99900 s",
        )
        self.assertEqual(
            CodeTimer._format_time(NS_PER_MINUTE, 7),
            "01:00.0000",
        )
        self.assertEqual(
            CodeTimer._format_time(NS_PER_HOUR, 7),
            "01:00:00.00",
        )
        self.assertEqual(
            CodeTimer._format_time(NS_PER_DAY, 7),
            "1 day, 00:00:00",
        )
        self.assertEqual(
            CodeTimer._format_time(2 * NS_PER_DAY, 7),
            "2 days, 00:00:00",
        )
        self.assertEqual(
            CodeTimer._format_time(12345678901234567890, 25),
            "142,889 days, 19:15:01.2345678900000",
        )

    @mock.patch("builtins.print")
    def test_properties(self, mock_print: mock.MagicMock):
        """Tests 'message' and 'time_str' properties."""
        # Test default values, if called before the measurement
        ct = CodeTimer()
        self.assertEqual(ct.message, "No code execution has taken place yet.")
        self.assertEqual(ct.time_str, "")
        # Test values after measurement correspond to the printout
        with ct:
            pass
        self.assertEqual(ct.message, mock_print.call_args.args[0])
        self.assertEqual(ct.time_str, re.search(r"took (.*)\.", ct.message).group(1))

    @mock.patch("builtins.print")
    def test_decorator(self, mock_print: mock.MagicMock):
        """Tests that the class can be used as a decorator."""

        @CodeTimer()
        def func():
            return 123

        self.assertEqual(func(), 123)
        mock_print.assert_called_once()
        self.assertIn("Code execution", mock_print.call_args.args[0])

    @flaky
    def test_results_collection(self):
        """Tests storing the benchmark results into a user-provided list."""
        results = []

        @CodeTimer(no_print=True, results=results)
        def sleep_ms(ms: int):
            sleep(ms / 1000)

        sleep_times = [10, 20, 30]
        for idx, t_sleep in enumerate(sleep_times):
            sleep_ms(t_sleep)
        for idx, t_sleep in enumerate(sleep_times):
            self.assertAlmostEqual(results[idx].ms, t_sleep, delta=2)


if __name__ == "__main__":
    unittest.main()



import pickle
from typing import Any


# ---------------------------------------------------------------------------
# FUNCTIONS
# ---------------------------------------------------------------------------
def pickle_object(obj: Any, filename: str, protocol=None) -> None:
    """
    Pickles the provided Python object and outputs it to the target file.
    """
    with open(filename, "wb") as f:
        pickle.dump(obj, f, protocol=protocol)


def unpickle_object(filename: str) -> Any:
    """
    Unpickles the Python object from a target file and returns it.
    """
    with open(filename, "rb") as f:
        return pickle.load(f)



import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

from jacktrade import merge_csv_files


class MergeCsvFilesTest(unittest.TestCase):
    """
    Tests merge_csv_files() function.

    NOTE: The behaviour of the function when has_headers=True but the source CSV
          files have mismatching headers is underfined and therefeore not tested.
    """

    SRC_FILES = ("tests/test_data/names_1.csv", "tests/test_data/names_2.csv")

    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.dest_file = str(Path(self.tempdir.name) / "merged.csv")

    def tearDown(self):
        self.tempdir.cleanup()

    def assert_file_contents_equal(self, file: str, contents: str) -> str:
        """Asserts that the file contains the provided text."""
        with open(file, "r") as f:
            self.assertEqual(f.read(), contents)

    def test_merge_has_headers(self):
        """Merge files considering the headers."""
        dest_file = merge_csv_files(
            self.SRC_FILES,
            self.dest_file,
            has_headers=True,
        )
        self.assert_file_contents_equal(
            dest_file,
            "NAME,SURNAME,AGE\nDon,Johnson,52\nWilliam,Andrews,27\nAdam,Smith,73\n",
        )

    def test_merge_no_headers(self):
        """Merge files' content verbatim when data is declared as not having headers."""
        dest_file = merge_csv_files(
            self.SRC_FILES,
            self.dest_file,
            has_headers=False,
        )
        self.assert_file_contents_equal(
            dest_file,
            "NAME,SURNAME,AGE\nDon,Johnson,52\nWilliam,Andrews,27\nAGE,SURNAME,NAME\n73,Smith,Adam\n",
        )

    def test_no_files_provided(self):
        """Function returns None when no merging took place."""
        self.assertIsNone(merge_csv_files([], self.dest_file))


if __name__ == "__main__":
    unittest.main()



import copy
import time
from functools import wraps
from math import log10
from typing import Callable

# ---------------------------------------------------------------------------
# CONSTANTS
# ---------------------------------------------------------------------------
NS_PER_DAY = 86400_000_000_000
NS_PER_HOUR = 3600_000_000_000
NS_PER_MINUTE = 60_000_000_000
NS_PER_SECOND = 1_000_000_000
NS_PER_MILLISECOND = 1_000_000
NS_PER_MICROSECOND = 1_000


# ---------------------------------------------------------------------------
# CLASSES
# ---------------------------------------------------------------------------
class CodeTimer:
    """
    Times a section of code inside a "with" statement.

    Example usage:
        ```py
        with CodeTimer() as t:
            # Code to time
        print(f"Execution took {t.s} seconds / {t.ms} milliseconds / {t.us} microseconds / {t.ns} nanoseconds.")
        ```

    Parameters:
        - no_print: If True, no message will be printed on exiting the timing block.
        - min_digits: Minimal number of digits to display when printing the result.
        - results:  A list where, if provided, CodeTimer instances holding timing results
                    are stored after every call. Useful when using the class as a decorator
                    to store the results of multiple wrapped function calls.

    Attributes:
        - ns: Code execution time in nanoseconds.
        - us: Code execution time in microseconds.
        - ms: Code execution time in milliseconds.
        - s: Code execution time in seconds.
        - m: Code execution time in minutes.
        - h: Code execution time in hours.
        - d: Code execution time in days.

    Properties:
        - time_str: Returns the formatted string representation of execution time.
        - message: Returns the message normally printed on measurement completion.
    """

    _UNITS_SECONDS = ("ns", "us", "ms", "s")

    def __init__(
        self, no_print: bool = False, min_digits: int = 3, results: list = None
    ):
        self._no_print = no_print
        self._min_digits = min_digits
        self._results = results
        self.ns = -1
        self.us = -1
        self.ms = -1
        self.s = -1
        self.m = -1
        self.h = -1
        self.d = -1

    @property
    def time_str(self) -> str:
        """
        Returns a formatted string representation of elapsed code execution time,
        where the unit and the number of decimal places are chosen automatically.

        Returns an empty string if no measurement has taken place yet.
        """
        return "" if (self.ns < 0) else self._format_time(self.ns, self._min_digits)

    @property
    def message(self) -> str:
        """
        Returns the message printed out at the end of code execution as a string.

        Returns a different message if no code execution has taken place yet.
        """
        if self.time_str:
            return f"Code execution took {self.time_str}."
        else:
            return "No code execution has taken place yet."

    def __enter__(self):
        self._start_time_ns = time.perf_counter_ns()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.ns = time_ns = time.perf_counter_ns() - self._start_time_ns
        self.us = time_ns / NS_PER_MICROSECOND
        self.ms = time_ns / NS_PER_MILLISECOND
        self.s = time_ns / NS_PER_SECOND
        self.m = time_ns / NS_PER_MINUTE
        self.h = time_ns / NS_PER_HOUR
        self.d = time_ns / NS_PER_DAY

        if not self._no_print:
            print(self.message)
        if self._results is not None:
            self._results.append(copy.copy(self))

    def __call__(self, function: Callable):
        """For using the class as a decorator."""

        @wraps(function)
        def wrapper(*args, **kwargs):
            with self:
                retval = function(*args, **kwargs)
            return retval

        return wrapper

    @classmethod
    def _format_time(cls, time_ns: int, min_digits: int) -> str:
        """Formats the time string based on the duration in nanoseconds."""
        # Sub 60 seconds, format in seconds, else use calendar time format
        if (time_ns / NS_PER_MINUTE) < 1:
            return cls._format_time_seconds(time_ns, min_digits)
        else:
            return cls._format_time_d_hh_mm_ss(time_ns, min_digits)

    @staticmethod
    def _count_digits(number: int) -> int:
        """Returns the number of digits in the provided integer."""
        # len(str(number)) is much slower than this solution
        return int(log10(number)) + 1 if number else 0

    @classmethod
    def _format_time_seconds(cls, time_ns: int, min_digits: int) -> str:
        """
        Automatically selects the suitable second-based time unit and returns
        a formatted string representation of time.
        """
        n_digits = cls._count_digits(time_ns)
        n_divisions = min(
            (n_digits - 1) // 3, len(cls._UNITS_SECONDS) - 1
        )  # How many times to divide by 1000
        digits_after_division = n_digits - 3 * n_divisions
        time_after_division = time_ns / (1000**n_divisions)
        decimal_places = max(0, min_digits - digits_after_division)
        fmt = (
            ",." + str(decimal_places) + "f"
        )  # How many digits spill over behind the decimal place
        return f"{time_after_division:{fmt}} {cls._UNITS_SECONDS[n_divisions]}"

    @classmethod
    def _format_time_d_hh_mm_ss(cls, time_ns: int, min_digits: int) -> str:
        """Formats the time in "[[d days, ]hh:]mm:ss[.ms]" format."""
        days, remainder_ns = divmod(time_ns, NS_PER_DAY)
        hours, remainder_ns = divmod(remainder_ns, NS_PER_HOUR)
        minutes, remainder_ns = divmod(remainder_ns, NS_PER_MINUTE)
        seconds, remainder_ns = divmod(remainder_ns, NS_PER_SECOND)
        # Determine the number of decimal places to display
        decimal_places = min_digits
        timestr = ""
        more_above = False  # Greater units of time are above zero
        if days > 0:
            decimal_places -= cls._count_digits(days)
            timestr += f"{days:,} day{'s' if days > 1 else ''}, "
            more_above = True
        if (time_ns / NS_PER_HOUR) >= 1:
            timestr += f"{hours:02d}:"
            decimal_places -= 2 if more_above else cls._count_digits(hours)
            more_above = True
        if (time_ns / NS_PER_MINUTE) >= 1:
            decimal_places -= 2 if more_above else cls._count_digits(minutes)
            more_above = True
        if (time_ns / NS_PER_SECOND) >= 1:
            decimal_places -= 2 if more_above else cls._count_digits(seconds)

        decimal_places = max(decimal_places, 0)
        return (
            timestr
            + f"{minutes:02d}:{seconds:02d}"
            + (
                f"{remainder_ns / NS_PER_SECOND:.{decimal_places}f}"[1:]
                if decimal_places
                else ""
            )
        )



import unittest
from dataclasses import dataclass
from typing import Iterator

from jacktrade import (
    BaseMapping,
    MasterDict,
    Permutations,
    chunkify,
    flatten_dict,
    flatten_list,
    get_first_dict_item,
    get_first_dict_key,
    get_first_dict_value,
    ichunkify,
    limit_iterator,
)

# ---------------------------------------------------------------------------
# TEST FIXTURES
# ---------------------------------------------------------------------------
MULTILEVEL_DICT = {"a1": {"a2": {"a3": 1, "b3": 2}, "b2": 3}, "b1": 4, "c1": 5}
MULTILEVEL_LIST = [[[1, 2], 3], 4, 5]
LONG_LIST = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]
CHUNKIFY_TEST_PARAMS = [
    # Parameters are (chunk size, first element, last element, element count)
    (1, [1], [20], 20),
    (3, [1, 2, 3], [19, 20], 7),
    (5, [1, 2, 3, 4, 5], [16, 17, 18, 19, 20], 4),
    (99, LONG_LIST, LONG_LIST, 1),
]


@dataclass
class Person:
    """For testing BaseMapping."""

    name: str
    age: int
    alive: bool


PEOPLE = (
    Person("John Doe", 27, True),
    Person("Jane Doe", 39, True),
    Person("David Hume", 313, False),
)

CORRUPT_PEOPLE = (Person("Zargothrax", "666", None), None)


# ---------------------------------------------------------------------------
# TEST CASES
# ---------------------------------------------------------------------------
class CollectionsTest(unittest.TestCase):
    """
    Tests the collections submodule.
    """

    def test_flatten_dict_core(self):
        """Tests flattenning a multilevel dict."""
        input_data = MULTILEVEL_DICT
        self.assertEqual(flatten_dict(input_data), [1, 2, 3, 4, 5])

    def test_flatten_dict_max_depth(self):
        """Tests flattening a multilevel dict when max_depth is reached."""
        input_data = MULTILEVEL_DICT
        self.assertEqual(
            flatten_dict(input_data, max_depth=2), [{"a3": 1, "b3": 2}, 3, 4, 5]
        )

    def test_get_first_dict_x(self):
        """Tests the get_first_dict_* functions."""
        test_dict = {"a": 1, "b": 2, "c": 3}
        self.assertEqual(get_first_dict_item(test_dict), ("a", 1))
        self.assertEqual(get_first_dict_key(test_dict), "a")
        self.assertEqual(get_first_dict_value(test_dict), 1)

    def test_flatten_list_core(self):
        """Tests flattening a multilevel list."""
        input_data = MULTILEVEL_LIST
        self.assertEqual(flatten_list(input_data), [1, 2, 3, 4, 5])

    def test_flatten_list_max_depth(self):
        """Tests flattening a multilevel list."""
        input_data = MULTILEVEL_LIST
        self.assertEqual(flatten_list(input_data, max_depth=2), [[1, 2], 3, 4, 5])

    def test_chunkify(self):
        """Tests chunkify function."""
        for chunk_size, first_exp, last_exp, count_exp in CHUNKIFY_TEST_PARAMS:
            with self.subTest(chunk_size=chunk_size):
                chunks = list(chunkify(LONG_LIST, chunk_size))
                self.assertEqual(first_exp, chunks[0])
                self.assertEqual(last_exp, chunks[-1])
                self.assertEqual(count_exp, len(chunks))

    def test_chunkify_infinite_chunk_size(self):
        """Tests chunkify when chunk_size is not specified, making it infinite."""
        self.assertEqual(list(chunkify(LONG_LIST, None)), [LONG_LIST])

    def test_ichunkify(self):
        """Tests ichunkify function."""
        for chunk_size, first_exp, last_exp, count_exp in CHUNKIFY_TEST_PARAMS:
            with self.subTest(chunk_size=chunk_size):
                # Chunks must be consumed as they are yielded
                for chunk_count, chunk in enumerate(
                    ichunkify(LONG_LIST, chunk_size), 1
                ):
                    self.assertIsInstance(chunk, Iterator)
                    chunk = list(chunk)  # Consume the chunk
                    if chunk_count == 1:
                        first_chunk = chunk
                last_chunk = chunk
                self.assertEqual(first_exp, first_chunk)
                self.assertEqual(last_exp, last_chunk)
                self.assertEqual(count_exp, chunk_count)

    def test_ichunkify_infinite_chunk_size(self):
        """Tests ichunkify when chunk_size is not specified, making it infinite."""
        self.assertEqual(list(next(ichunkify(LONG_LIST, None))), LONG_LIST)

    def test_limited_iterator(self):
        """Tests limiting the iterator to max elements."""
        self.assertEqual(list(limit_iterator(LONG_LIST, 5)), [1, 2, 3, 4, 5])
        self.assertEqual(list(limit_iterator(iter(LONG_LIST), 5)), [1, 2, 3, 4, 5])
        self.assertEqual(list(limit_iterator(LONG_LIST, 99)), LONG_LIST)
        self.assertEqual(list(limit_iterator(iter(LONG_LIST), 99)), LONG_LIST)

    def test_permutations(self):
        """Tests Permutations class."""
        kwargs_in = {"n": [1, 2, 3], "c": ["A", "B"]}
        perms = Permutations(**kwargs_in)
        kwargs_out = [
            {"n": 1, "c": "A"},
            {"n": 1, "c": "B"},
            {"n": 2, "c": "A"},
            {"n": 2, "c": "B"},
            {"n": 3, "c": "A"},
            {"n": 3, "c": "B"},
        ]
        self.assertEqual(len(perms), 6)
        self.assertEqual(
            perms.args, [(1, "A"), (1, "B"), (2, "A"), (2, "B"), (3, "A"), (3, "B")]
        )
        self.assertEqual(perms.kwargs, kwargs_out)
        self.assertEqual(list(perms), kwargs_out)  # Tests __iter__
        self.assertIn(str(kwargs_in), str(perms))  # Tests __repr__
        self.assertIn(perms.__class__.__name__, str(perms))  # Tests __repr__


class MasterDictTest(unittest.TestCase):
    """
    Tests MasteDict class.
    """

    def setUp(self) -> None:
        self.master = MasterDict(
            letters={1: "A", 2: "B"}, numbers={1: "1", 2: "2", 3: "3"}
        )

    def test_delete_keys(self):
        """Tests deleting individual keys from a collection of dicts."""
        self.master.delete_keys(1, 3)
        self.assertEqual(self.master.letters, {2: "B"})  # Deleted: 1
        self.assertEqual(self.master.numbers, {2: "2"})  # Deleted: 1, 3

    def test_clear_all(self):
        """Tests clearing all sub-dicts."""
        self.master.clear_all()
        self.assertEqual(self.master.letters, {})
        self.assertEqual(self.master.numbers, {})

    def test_as_dict(self):
        """Tests converting a MasterDict instance to a dict."""
        self.assertEqual(
            self.master.as_dict(),
            {"letters": {1: "A", 2: "B"}, "numbers": {1: "1", 2: "2", 3: "3"}},
        )

    def test_str_and_repr(self):
        """Tests string representations of a MasterDict object."""
        display = (
            "MasterDict(letters={1: 'A', 2: 'B'}, numbers={1: '1', 2: '2', 3: '3'})"
        )
        self.assertEqual(repr(self.master), display)
        self.assertEqual(str(self.master), display)


class BaseMappingTest(unittest.TestCase):
    """
    Tests BaseMapping class.
    """

    @staticmethod
    def first_name_getter(person: Person) -> str:
        return person.name.split(" ")[0]

    @staticmethod
    def age_getter(person: Person) -> int:
        return person.age

    @staticmethod
    def is_alive(person: Person) -> bool:
        return person.alive is True

    @staticmethod
    def is_less_than_30(person: Person) -> bool:
        return person.age < 30

    def test_basic_mapping(self):
        """Tests basic mapping with key and value getters."""
        mapping = BaseMapping(PEOPLE, self.first_name_getter, self.age_getter)
        expected_result = {"John": 27, "Jane": 39, "David": 313}
        self.assertEqual(mapping, expected_result)

    def test_with_condition(self):
        """Tests mapping with a condition function provided."""
        mapping = BaseMapping(
            PEOPLE, self.first_name_getter, self.age_getter, self.is_less_than_30
        )
        expected_result = {"John": 27}
        self.assertEqual(mapping, expected_result)
        mapping = BaseMapping(
            PEOPLE, self.first_name_getter, self.age_getter, self.is_alive
        )
        expected_result = {"John": 27, "Jane": 39}
        self.assertEqual(mapping, expected_result)

    def test_skip_exceptions(self):
        """Tests discarding elements which trigger an exception."""
        mapping = BaseMapping(
            PEOPLE + CORRUPT_PEOPLE,
            self.first_name_getter,
            self.age_getter,
            self.is_less_than_30,
            skip_exceptions=(TypeError, AttributeError),
        )
        expected_result = {"John": 27}
        self.assertEqual(mapping, expected_result)

    def test_raise_exceptions(self):
        """Exception is raised because not all errors are handled."""
        # Skip no exceptions
        with self.assertRaises(TypeError):
            mapping = BaseMapping(
                PEOPLE + CORRUPT_PEOPLE,
                self.first_name_getter,  # Raises AttributeError (2nd)
                self.age_getter,
                self.is_less_than_30,  # Raises TypeError (1st)
            )
        # Skip first exception but not the second one
        with self.assertRaises(AttributeError):
            mapping = BaseMapping(
                PEOPLE + CORRUPT_PEOPLE,
                self.first_name_getter,  # Raises AttributeError (2nd)
                self.age_getter,
                self.is_less_than_30,  # Raises TypeError (1st)
                skip_exceptions=(TypeError,),
            )

    def test_repr(self):
        """Tests that __repr__ magic method displays the name of the derived class."""

        class DerivedMapping(BaseMapping):
            def __init__(self, items):
                # This is not how one would usually use a class
                super().__init__(items, lambda x: x[0], lambda x: x[1])

        mapping = repr(DerivedMapping([("Will", 29)]))
        # Derived class name is displated first
        self.assertTrue(mapping.startswith("DerivedMapping"))
        # Contents are also displayed, like in a normal dict
        self.assertIn(repr({"Will": 29}), mapping)

    def test_invert(self):
        """Tests inverting the mapping."""
        mapping = BaseMapping([(1, "a"), (2, "b")], lambda x: x[0], lambda x: x[1])
        self.assertEqual(mapping.invert(), {"a": 1, "b": 2})

    def test_invert_strict_mode(self):
        """Tests strict mode when inverting the mapping."""
        mapping = BaseMapping([(1, "a"), (2, "a")], lambda x: x[0], lambda x: x[1])
        self.assertEqual(len(mapping.invert(strict=False)), 1)
        with self.assertRaises(ValueError):
            mapping.invert(strict=True)

    def test__invert_and_cast(self):
        """Tests inverting the mapping and casting it to another mapping type."""
        first_name_getter = self.first_name_getter
        age_getter = self.age_getter

        class NameToAge(BaseMapping):
            def __init__(self, items):
                super().__init__(items, first_name_getter, age_getter)

            def invert(self, strict: bool = False) -> "AgeToName":
                return super()._invert_and_cast(AgeToName, strict)

        class AgeToName(BaseMapping):
            def __init__(self, items):
                super().__init__(items, age_getter, first_name_getter)

            def invert(self, strict: bool = False) -> NameToAge:
                return self._invert_and_cast(NameToAge, strict)

        mapping = AgeToName(PEOPLE)
        inverse_mapping = mapping.invert()
        self.assertIsInstance(inverse_mapping, NameToAge)
        self.assertEqual(mapping, inverse_mapping.invert())

    def test__invert_and_cast_to_generic_type(self):
        """Tests inverting the mapping and casting it to a generic dict type."""
        first_name_getter = self.first_name_getter
        age_getter = self.age_getter

        class AgeToName(BaseMapping):
            def __init__(self, items):
                super().__init__(items, age_getter, first_name_getter)

            def invert(self, strict: bool = False) -> dict:
                return self._invert_and_cast(dict, strict)

        mapping = AgeToName(PEOPLE)
        inverse_mapping = mapping.invert()
        self.assertEqual(type(inverse_mapping), dict)
        self.assertEqual(mapping, BaseMapping.invert(inverse_mapping))


if __name__ == "__main__":
    unittest.main()



import platform
import subprocess
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import Mock, patch

from jacktrade import hibernate, in_virtual_environment, restart, shutdown, suspend

PYTHON_CMDS = "import os;import sys;sys.path.insert(0, os.getcwd());from jacktrade import in_virtual_environment;print(in_virtual_environment(),end='');"


class SysenvTest(unittest.TestCase):
    """
    Tests the sysenv module.
    """

    def test_in_virtual_environment(self):
        """
        Tests detecting whether the code is running from a virtual environment.
        """
        in_virtual_environment()  # Hit for code coverage

        with TemporaryDirectory() as tmpd:
            venv_path = Path(tmpd).resolve() / "venv"
            operating_system = platform.system()
            if operating_system == "Windows":
                python = "python"
                sep = " & "
                test_cmd = f'"{venv_path}\\Scripts\\python.exe" -c "{PYTHON_CMDS}"'
            elif operating_system in ("Linux", "Darwin"):
                python = "python3"
                sep = "; "
                test_cmd = f'"{venv_path}/bin/python" -c "{PYTHON_CMDS}"'
            else:
                raise NotImplementedError("Operating system not supported.")
            # Running outside of venv (will not work if launched from venv)
            self.assertEqual(
                subprocess.check_output(
                    f'{python} -c "{PYTHON_CMDS}"', shell=True
                ).decode(),
                "False",
            )
            # Running in venv
            commands = sep.join(
                [
                    f'{python} -m venv "{venv_path}"',
                    test_cmd,
                ]
            )
            self.assertEqual(
                subprocess.check_output(commands, shell=True).decode(), "True"
            )


@patch("subprocess.run")
class PowerManagementTest(unittest.TestCase):
    """
    Tests power management functions.

    Note that, for obvious reasons, these tests do not validate that
    the intended side effect actually occurs, only that a function
    calls a particular command under particular circumstances.
    """

    FUNCTIONS = (suspend, hibernate, shutdown, restart)

    @patch("platform.system", return_value="Windows")
    def test_power_management_windows(self, mock_system: Mock, mock_run: Mock):
        """Tests power management calls on Windows."""
        cmds = (
            "Rundll32.exe Powrprof.dll,SetSuspendState Sleep",
            "shutdown /h",
            "shutdown /s",
            "shutdown /r",
        )
        for func, cmd in zip(self.FUNCTIONS, cmds):
            func()
            mock_run.assert_called_with(cmd)

    @patch("platform.system", return_value="Linux")
    def test_power_management_linux(self, mock_system: Mock, mock_run: Mock):
        cmds = (
            "systemctl suspend",
            "systemctl hibernate",
            "shutdown -h now",
            "shutdown -r now",
        )
        for func, cmd in zip(self.FUNCTIONS, cmds):
            func()
            mock_run.assert_called_with(cmd, shell=True)

    @patch("platform.system", return_value="Mythical")
    def test_power_management_others(self, mock_system: Mock, mock_run: Mock):
        for func in self.FUNCTIONS:
            with self.assertRaisesRegex(NotImplementedError, "Mythical"):
                func()


if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

from jacktrade import StringBuffers

FILE_1 = "abc.txt"
FILE_2 = "def.txt"
TEST_DATA = [
    (FILE_1, "1"),
    (FILE_2, "2"),
    (FILE_1, "3"),
    (FILE_1, "4"),
]  # FILE_1: "134", FILE_2: "2"


class StringBuffersTest(unittest.TestCase):
    """
    Tests the StringBuffers class.
    """

    def setUp(self) -> None:
        self.tempdir = TemporaryDirectory()
        self.file_1_path = Path(self.tempdir.name) / FILE_1
        self.file_2_path = Path(self.tempdir.name) / FILE_2

    def tearDown(self) -> None:
        self.tempdir.cleanup()

    def load_data(self, buffers: StringBuffers) -> StringBuffers:
        """Loads data for test cases."""
        for args in TEST_DATA:
            buffers.add(*args)
        return buffers

    def assert_in_file(self, file_path: str, file_content: str):
        """Checks the file's contents."""
        with open(file_path) as f:
            self.assertEqual(f.read(), file_content)

    def test_auto_flush(self):
        """Tests that buffer_size can be None."""
        self.load_data(StringBuffers(self.tempdir.name, buffer_size=2))
        self.assert_in_file(self.file_1_path, "13")
        self.assertFalse(self.file_2_path.exists())

    def test_flush(self):
        buffers = self.load_data(StringBuffers(self.tempdir.name))
        self.assertFalse(self.file_1_path.exists())
        self.assertFalse(self.file_2_path.exists())
        buffers.flush(FILE_1)
        self.assert_in_file(self.file_1_path, "134")
        self.assertFalse(self.file_2_path.exists())
        self.assertEqual(buffers.files, [FILE_1, FILE_2])

    def test_flush_nonexistent_file(self):
        buffers = self.load_data(StringBuffers(self.tempdir.name))
        nonexistent_filename = "totally_not_an_existing_file.txt"
        buffers.flush(nonexistent_filename)  # Passes if no exception is raised

    def test_flush_all(self):
        buffers = self.load_data(StringBuffers(self.tempdir.name))
        self.assertFalse(self.file_1_path.exists())
        self.assertFalse(self.file_2_path.exists())
        buffers.flush_all()
        self.assert_in_file(self.file_1_path, "134")
        self.assert_in_file(self.file_2_path, "2")
        self.assertEqual(buffers.files, [FILE_1, FILE_2])

    def test_remove(self):
        buffers = self.load_data(StringBuffers(self.tempdir.name))
        self.assertEqual(buffers.files, [FILE_1, FILE_2])
        self.assertFalse(self.file_1_path.exists())
        self.assertFalse(self.file_2_path.exists())
        # Dump and remove file 1
        buffers.remove(FILE_1)
        self.assert_in_file(self.file_1_path, "134")
        self.assertFalse(self.file_2_path.exists())  # File 2 has not been dumped yet
        self.assertEqual(buffers.files, [FILE_2])
        # Dumpr and remove file 2
        buffers.remove(FILE_2)
        self.assert_in_file(self.file_2_path, "2")
        self.assertEqual(buffers.files, [])

    def test_remove_all(self):
        buffers = self.load_data(StringBuffers(self.tempdir.name))
        self.assertFalse(self.file_1_path.exists())
        self.assertFalse(self.file_2_path.exists())
        removed_files = buffers.files  # All fiiles
        self.assertEqual(buffers.remove_all(), removed_files)
        self.assert_in_file(self.file_1_path, "134")
        self.assert_in_file(self.file_2_path, "2")
        self.assertEqual(buffers.files, [])

    def test_files_property(self):
        buffers = self.load_data(StringBuffers(self.tempdir.name))
        self.assertEqual(buffers.files, [FILE_1, FILE_2])

    def test_iter(self):
        buffers = self.load_data(StringBuffers(self.tempdir.name))
        iterable = list(buffers)
        self.assertEqual(iterable[0][0], FILE_1)
        self.assertEqual(list(iterable[0][1]), ["1", "3", "4"])
        self.assertEqual(iterable[1][0], FILE_2)
        self.assertEqual(list(iterable[1][1]), ["2"])


if __name__ == "__main__":
    unittest.main()



# Because each submodule is small, there is no need to further subdivide imports
# into categories. Therefore, make everything available from the top-level module,
# like a Swiss army knife.
from .benchmark import CodeTimer
from .buffers import StringBuffers
from .collections import (
    BaseMapping,
    MasterDict,
    Permutations,
    chunkify,
    flatten_dict,
    flatten_list,
    get_first_dict_item,
    get_first_dict_key,
    get_first_dict_value,
    ichunkify,
    limit_iterator,
)
from .files import merge_csv_files
from .multicore import do_multicore_work
from .pickler import pickle_object, unpickle_object
from .sysenv import hibernate, in_virtual_environment, restart, shutdown, suspend



import csv


# ---------------------------------------------------------------------------
# FUNCTIONS
# ---------------------------------------------------------------------------
def merge_csv_files(
    src_files: list[str], dest_file: str, has_headers: bool = True
) -> str | None:
    """
    Merges multiple CSV files into a single one.
    Parameters:
        - src_files: A list of paths to source files to combine.
        - dest_file: Path to the output file.
        - has_headers:
            If True (default), the first row of each CSV files is treated as a header
            and is written to the output file only once. The files must have the identical
            number of columns and column names, but not necessarily in the same order.
            If False, CSV files' contents are concatenated in full to one another.

    Returns:
        - Path to the output file if the file has been created.
        - None if the output file nas not been created.
    """
    if not src_files:
        return None  # Don't even create a new file if there are no sources
    with open(dest_file, "w", newline="") as fd:
        if has_headers:
            # Peak into the first file to obtain headers
            with open(src_files[0], "r") as fs:
                reader = csv.DictReader(fs)
                column_names = reader.fieldnames
            # Initialise the writer with headers
            writer = csv.DictWriter(fd, column_names)
            writer.writeheader()
        else:
            writer = csv.writer(fd)
        # Iterate over the files and write their contents into dest
        for src_file in src_files:
            with open(src_file, "r") as fs:
                reader_type = csv.DictReader if has_headers else csv.reader
                writer.writerows(reader_type(fs))
    return dest_file



import platform
import subprocess
import sys


def in_virtual_environment() -> bool:
    """
    Returns True if called from a virtual environment, else False.
    """
    # https://docs.python.org/3/library/venv.html#how-venvs-work
    # sys.base_prefix: Always points to the Python installation directory.
    # sys.prefix: Points to either venv or installation directory.
    return sys.prefix != sys.base_prefix


def _execute_os_specific_commands(windows_cmd: str = None, linux_cmd: str = None):
    """Executes OS-specific commands depending on the current platform."""
    match platform.system():
        case "Windows" if windows_cmd:
            subprocess.run(windows_cmd)
        case "Linux" if linux_cmd:
            subprocess.run(linux_cmd, shell=True)
        case str(p):
            raise NotImplementedError(f"System '{p}' is not supported.")


def suspend() -> None:
    """
    Puts the machine into standby. "sleep" on Windows and "suspend" on Linux.
    """
    _execute_os_specific_commands(
        "Rundll32.exe Powrprof.dll,SetSuspendState Sleep", "systemctl suspend"
    )


def hibernate() -> None:
    """Puts the machine into hibernation."""
    _execute_os_specific_commands("shutdown /h", "systemctl hibernate")


def shutdown() -> None:
    """Shuts down the machine."""
    _execute_os_specific_commands("shutdown /s", "shutdown -h now")


def restart() -> None:
    """Restarts the machine."""
    _execute_os_specific_commands("shutdown /r", "shutdown -r now")



from itertools import chain, islice, product
from typing import Any, Callable, Hashable, Iterable, Iterator, Optional, TypeVar


# ---------------------------------------------------------------------------
# DICTIONARIES
# ---------------------------------------------------------------------------
def _flatten_dict(input_data: dict, output_data: list, max_depth: int) -> None:
    """
    Flattens a nested dict and returns the values inside output_data list.
    Recursion happens until a non-dict item is encountered, or max_depth is reached.
    """
    if isinstance(input_data, dict) and (max_depth > 0):
        for data in input_data.values():
            _flatten_dict(data, output_data, max_depth - 1)
    else:
        output_data.append(input_data)


def flatten_dict(input_data: dict, max_depth: int = 999) -> list:
    """
    Flattens a nested dict and returns the values inside a list.
    Recursion happens until a non-dict item is encountered, or max_depth is reached.
    """
    output_data = []
    _flatten_dict(input_data, output_data, max_depth)
    return output_data


def get_first_dict_item(dictionary: dict) -> tuple[Any, Any]:
    """Returns the first (key, value) pair a dictionary."""
    return next(iter(dictionary.items()), None)


def get_first_dict_key(dictionary: dict) -> Any:
    """Returns the first key in a dictionary."""
    return next(iter(dictionary.keys()), None)


def get_first_dict_value(dictionary: dict) -> Any:
    """Returns the first value in a dictionary."""
    return next(iter(dictionary.values()), None)


class MasterDict:
    """
    Holds multiple dicts and provides a simple method to delete keys and values from them all.

    This object is usually used to hold caches, which can then be reliably emptied with
    a single method call to prevent memory leaks.

    Despite the name, this class is not a dict subtype. However, it can be converted to a
    dict using as_dict() method.
    """

    def __init__(self, **subdicts: dict) -> None:
        """
        Initialises the master dict with sub-dicts, which are provided as keyword
        arguments. Argument name is the dict name, while value is the dict object.
        """
        for attr, value in subdicts.items():
            setattr(self, attr, value)

    def delete_keys(self, *keys: Hashable) -> None:
        """Deletes provided keys from all sub-dictionaries where they are present."""
        for key in keys:
            for subdict in self:
                subdict.pop(key, None)

    def clear_all(self) -> None:
        """Clears all sub-dicts."""
        for subdict in self:
            subdict.clear()

    def as_dict(self) -> dict[str, dict]:
        """
        Returns a dictionary representation of itself, where keys are sub-dict names and
        values are sub-dict objects.
        """
        return vars(self)

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + "("
            + ", ".join(f"{n}={d}" for n, d in vars(self).items())
            + ")"
        )

    def __iter__(self) -> Iterator[dict]:
        """Iterates over sub-dicts."""
        return iter(vars(self).values())


# ---------------------------------------------------------------------------
# ITERABLES
# ---------------------------------------------------------------------------
def _flatten_list(input_data: list, output_data: list, max_depth: int) -> None:
    """
    Flattens a multilevel list and returns the values inside output_data list.
    Recursion happens until a non-list item is encountered, or max_depth is reached.
    """
    if isinstance(input_data, list) and (max_depth > 0):
        for data in input_data:
            _flatten_list(data, output_data, max_depth - 1)
    else:
        output_data.append(input_data)


def flatten_list(input_data: list, max_depth: int = 999) -> list:
    """
    Flattens a multilevel list and returns the values inside a list.
    Recursion happens until a non-list item is encountered, or max_depth is reached.
    """
    output_data = []
    _flatten_list(input_data, output_data, max_depth)
    return output_data


def chunkify(iterable: Iterable, chunk_size: int = None) -> Iterator[list]:
    """
    Yields successive n-sized list chunks from an iterable.
    Supports generator expressions.
    """
    iterator = iter(iterable)  # Must be assigned here, else infinite loop
    while chunk := list(islice(iterator, chunk_size)):
        yield chunk


def ichunkify(iterable: Iterable, chunk_size: int = None) -> Iterator[Iterator]:
    """
    Yields successive n-sized iterator chunks from an iterable.
    Supports generator expressions.

    WARNING!:   Chunks must be consumed as they are yielded, otherwise
                the results will not be as expected!
    """
    iterator = iter(iterable)
    chunk_size_minus_one = None if (chunk_size is None) else (chunk_size - 1)
    for first_chunk_elem in iterator:
        yield chain([first_chunk_elem], islice(iterator, chunk_size_minus_one))


def limit_iterator(iterable: Iterable, limit: int = None) -> Iterator:
    """
    Returns an interator which yields successive elements of
    the iterable up to the limit, if the limit is specified.
    """
    for count, item in enumerate(iterable, 1):
        if limit and (count > limit):
            return
        yield item


# ---------------------------------------------------------------------------
# MAPPINGS
# ---------------------------------------------------------------------------
T = TypeVar("T")


class BaseMapping(dict):
    """
    Base mapping where each key and value are obtained by applying a getter to
    each item in the iterable.
    """

    def __init__(
        self,
        items: Iterable[T],
        key_getter: Callable[[T], Hashable],
        value_getter: Callable[[T], Any],
        condition: Optional[Callable[[T], bool]] = None,
        skip_exceptions: tuple[type[Exception]] = (),
    ) -> None:
        """
        Parameters:
            - items: An iterable of items of the same type to create a mapping from.
            - key_getter: A function accepting an item and returning a hashable dict key.
            - value_getter: A function accepting an item and returning a dict value.
            - condition: A function accepting an item and returning a boolean value
                         indicating if the item should be included in the mapping.
            - skip_exceptions: A tuple of exceptions which, if they appear during the
                               mapping construction, discard that item instead of raising.
        """
        for item in items:
            try:
                if condition is None or condition(item):
                    self[key_getter(item)] = value_getter(item)
            except skip_exceptions:
                pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict(self)})"

    def invert(self, strict: bool = False) -> dict:
        """
        Returns an inverse mapping of values to keys.

        When strict=True, ValueError is raised if the values cannot be unambiguously
        mapped to keys because they are not unique.
        """
        if strict and len(set(self.values())) != len(self):
            raise ValueError("Dict values are not unique.")
        return {v: k for k, v in self.items()}

    def _invert_and_cast(self, new_type: type[dict], strict: bool) -> dict:
        """
        Inverts the keys and values in the dictionary and returns them as a custom mapping type.

        This method is generally reserved for use with BaseMapping subclasses, when the inverse
        BaseMapping subclass instance is another BaseMapping subclass instance.
        """
        inverse_mapping = new_type([])
        # BaseMapping must be provided explicitly here to avoid infinite recursion
        inverse_mapping.update(BaseMapping.invert(self, strict))
        return inverse_mapping


# ---------------------------------------------------------------------------
# COMBINATORICS
# ---------------------------------------------------------------------------
class Permutations:
    """
    Yields all possible combinations of the named input parameters,
    returning them as positional (tuple) or keyword (dict) arguments.

    For example:
    ```py
    p = Permutations(a=[1, 2, 3], b=["A", "B"])
    for kwargs in p:
        # yields:
        # {"a": 1, "b": "A"}
        # {"a": 1, "b": "B"}
        # {"a": 2, "b": "A"}
        # {"a": 2, "b": "B"}
        # ...
    ```
    """

    def __init__(self, **kwargs: Iterable):
        self._kwargs = kwargs
        self._names = tuple(kwargs.keys())
        self._combinations = tuple(product(*kwargs.values()))

    @property
    def args(self) -> list[tuple]:
        """Lists all combinations as tuples (positional arguments)."""
        return list(self._combinations)

    @property
    def kwargs(self) -> list[dict[str, Any]]:
        """Lists all combinations as dicts (keyword arguments)."""
        return list(self)

    def __iter__(self) -> Iterator[dict[str, Any]]:
        """
        Yields a kwarg dictionary with all possible combinations of input parameters.
        """
        for values in self._combinations:
            yield dict(zip(self._names, values))

    def __len__(self) -> int:
        return len(self._combinations)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._kwargs}"



import unittest
from multiprocessing import cpu_count
from unittest import mock

from jacktrade import do_multicore_work

# ---------------------------------------------------------------------------
# TEST FIXTURES
# ---------------------------------------------------------------------------
NUMBERS = [1, 2, 3, 4, 5]
LETTERS = ["a", "b", "c", "d", "e"]
COMBINED = list(zip(NUMBERS, LETTERS))


def worker(first, second) -> tuple:
    """
    Sample worker to be deployed on multiple cores.
    Receives two arguments and returns them as a tuple.
    """
    return (first, second)


# ---------------------------------------------------------------------------
# TEST CASES
# ---------------------------------------------------------------------------
class MulticoreTest(unittest.TestCase):
    """
    Tests the multicore submodule.
    """

    def setUp(self) -> None:
        self.results = []

    def worker_done_callback(self, future):
        """Called when a worker finishes execution."""
        self.results.append(future.result())

    def test_no_args_or_kwargs_provided(self):
        """Tests that the function exits gracefully if no args or kwargs are provided."""
        do_multicore_work(worker)  # if It doesn't raise an exception, the test passes

    def test_args_provided(self):
        """Tests providing positional arguments to the worker."""
        do_multicore_work(
            worker,
            args=zip(NUMBERS, LETTERS),
            worker_done_callback=self.worker_done_callback,
        )
        self.assertEqual(set(self.results), set(COMBINED))

    def test_kwargs_provided(self):
        """Tests providing positional arguments to the worker."""
        do_multicore_work(
            worker,
            kwargs=[{"first": x, "second": y} for x, y in zip(NUMBERS, LETTERS)],
            worker_done_callback=self.worker_done_callback,
        )
        self.assertEqual(set(self.results), set(COMBINED))

    def test_args_and_kwargs_provided(self):
        """Tests providing positional arguments to the worker."""
        do_multicore_work(
            worker,
            args=zip(NUMBERS),  # zip will "tuplify" the list's elements
            kwargs=[{"second": y} for y in LETTERS],
            worker_done_callback=self.worker_done_callback,
        )
        self.assertEqual(set(self.results), set(COMBINED))

    @mock.patch("concurrent.futures.as_completed")
    @mock.patch("concurrent.futures.ProcessPoolExecutor")
    def test_idle_cpus(self, mock_executor, mock_as_completed):
        """Tests the use of idle_cpus parameter."""
        # Test params are (idle_cpus, max_workers)
        test_params = [(1, max(cpu_count() - 1, 1)), (cpu_count() + 1, 1)]
        for idle_cpus, max_workers in test_params:
            with self.subTest(idle_cpus=idle_cpus, max_workers=max_workers):
                mock_as_completed.return_value = []
                do_multicore_work(
                    worker, args=zip(NUMBERS, LETTERS), idle_cpus=idle_cpus
                )
                mock_executor.assert_called_with(max_workers)


if __name__ == "__main__":
    unittest.main()



from collections import defaultdict, deque
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterator


# ---------------------------------------------------------------------------
# CLASSES
# ---------------------------------------------------------------------------
class StringBuffers:
    """
    Sorts, stores and flushes string data to a disk. The main purpose
    of this buffer is to reduce the number of file open/close cycles
    and speed up logging of string data to disk.
    """

    def __init__(self, output_dir: str = ".", buffer_size: int = None):
        self._output_dir = Path(output_dir)
        self._output_dir.mkdir(parents=True, exist_ok=True)
        self._buffer_size = buffer_size
        self._buffers = defaultdict(lambda sz=buffer_size: deque(maxlen=sz))

    def _flush_buffer(self, output_file: str, buffer: deque) -> str:
        """
        Flushes the buffer to the output file.
        Moved to a separate function for DRY, and so that flush_all could
        iterate over the dict more efficiently.
        Returns the full path to the file to which the data was flushed.
        """
        output_file_path = self._output_dir / output_file
        try:
            # Try getting the first item. If there isn't one,
            # no point to even open the file.
            first_item = buffer.popleft()
            with open(output_file_path, "a") as f:
                f.write(first_item)
                while True:
                    f.write(buffer.popleft())
        except IndexError:
            pass  # When the buffer is empty, flushing is done
        return str(output_file_path)

    def add(self, output_file: str, data: str) -> int:
        """
        Adds the data to the buffer for the specified output file.
        Returns the number of elements in the buffer after the addition.
        If the addition causes the buffer to reach its capacity, the buffer
        is automatically flushed to the disk.
        """
        (buffer := self._buffers[output_file]).append(data)
        if self._buffer_size and (len(buffer) >= self._buffer_size):
            self._flush_buffer(output_file, buffer)
        return len(buffer)

    def flush(self, output_file: str) -> str:
        """
        Flushes the data in output file's buffer to the output file.
        Returns the full path to the file to which the data was flushed.
        """
        return self._flush_buffer(output_file, self._buffers[output_file])

    def flush_all(self) -> None:
        """Flushes all buffers to disk."""
        for output_file, buffer in self._buffers.items():
            self._flush_buffer(output_file, buffer)

    def remove(self, output_file: str) -> str:
        """
        Removes the output file from the buffer,
        flushing data to disk if it is not empty.
        Returns the full path to the file to which the data was flushed.
        """
        output_file_path = self.flush(output_file)
        del self._buffers[output_file]
        return output_file_path

    def remove_all(self) -> list[str]:
        """
        Removes all output files from the buffer,
        flushing the data to disk beforehand.
        Returns a list of paths to all removed files.
        """
        self.flush_all()  # Faster than doing an element-wise lookup
        removed_files = self.files
        self._buffers.clear()  # Batch-delete all buffers
        return removed_files

    @property
    def files(self) -> list[str]:
        """
        Returns a list of output files bound to string buffers.
        """
        return list(self._buffers.keys())

    def __iter__(self) -> Iterator[tuple]:
        return iter(self._buffers.items())


@dataclass
class TimestampedElement:
    timestamp: float
    data: Any


class TimedBuffer:
    """
    A rotating buffer holding timestamped elements.

    Parameters:
        - duration: When a buffer is cleaned, elements older than this value are discarded.
        - size: Maximum number of elements the buffer holds. If None
                (default), the buffer's size is unlimited.
    """

    def __init__(self, duration: float, size: int | None = None):
        if duration <= 0:
            raise ValueError("'duration' must be positive.")
        self._duration = duration
        self._buffer = deque(maxlen=size)

    def add(self, element: TimestampedElement) -> None:
        """Adds a timestamped element to the buffer."""
        self._buffer.append(element)

    def clean(self, now: float) -> None:
        """
        Discards all expired elements, meaning they are older
        than the buffer's duration parameter.

        Note: An element where 'timestamp == now - duration'
        is also discarded. This is because it is assumed that,
        from the moment this method was called, an infinitesimaly small
        amount of time has passed, and hence the element is outdated by
        the time it is processed.
        """
        try:
            # Keep removing elements from the start until
            # finding one that is inside the time interval.
            starting_point = now - self._duration
            while True:
                element = self._buffer.popleft()
                if element.timestamp > starting_point:
                    # First element within time constraints found
                    # Put it back in the buffer and exit
                    self._buffer.appendleft(element)
                    return
        except IndexError:
            return

    def __iter__(self) -> Iterator:
        return iter(self._buffer)


if __name__ == "__main__":
    tb = TimedBuffer(5)
    for i in range(20):
        tb.add(TimestampedElement(i, i))
    print(list(tb))
    tb.clean(22)
    print(list(tb))



import os.path
import tempfile
import unittest

from jacktrade import pickle_object, unpickle_object


class PickerTest(unittest.TestCase):
    """
    Tests the convenience pickling functions.
    """

    def test_pickle_unpickle(self):
        """Tests pickling and unpickling an object."""
        with tempfile.TemporaryDirectory() as td:
            obj = [1, 2, 3]
            filename = os.path.join(td, "obj.pickle")
            pickle_object(obj, filename)
            self.assertEqual(unpickle_object(filename), obj)


if __name__ == "__main__":
    unittest.main()



import concurrent.futures
import multiprocessing as mp
from itertools import zip_longest
from typing import Any, Callable, Iterable


# ---------------------------------------------------------------------------
# FUNCTIONS
# ---------------------------------------------------------------------------
def do_multicore_work(
    worker: Callable,
    args: Iterable[tuple] = None,
    kwargs: Iterable[dict] = None,
    worker_done_callback: Callable[[concurrent.futures.Future], Any] = None,
    idle_cpus: int = 0,
):
    """
    Splits the work done by the worker function across multiple CPU cores in a way which
    does not leak memory and uses only the resources required for the active workers.

    WARNING: This function must be run inside 'if __name__ == "__main__":' construct!

    Parameters:
        - worker: A function which does work.
        - args: An iterable containing tuples of positional arguments.
        - kwargs: An iterable containing dictionaries of keyword arguments.
        - worker_done_callback: A function which is called with a Future object as the only
                                parameter as soon as the worker completes assigned work.
        - idle_cpus: How many CPU cores to leave unoccupied. At minimum 1 core will be used.


    NOTE:   This function assumes that the task the worker executes is CPU bound and that
            the operating system will assign each worker to a different core, achieving
            optimal resource usage. However, there is no guarantee this will happen as
            CPU affinity is not explicitly set.
    """
    if not any((args, kwargs)):  # Both are None or empty iterables
        return  # Workers cannot work without arguments
    elif all((args, kwargs)):  # Both are provided
        # Use zip to exit early if iterables do not have the same length
        # and one of them runs out before the other
        workload = zip(args, kwargs)
    else:  # Either args or kwargs are provided, but not both
        workload = zip_longest(args or [], kwargs or [], fillvalue={} if args else ())

    max_workers = max(mp.cpu_count() - idle_cpus, 1)

    with concurrent.futures.ProcessPoolExecutor(max_workers) as executor:
        futures = (executor.submit(worker, *a, **ka) for a, ka in workload)
        for future in concurrent.futures.as_completed(futures):
            if worker_done_callback:
                worker_done_callback(future)
